import logging
import logging.config
from pathlib import Path
from typing import Any, Dict, Optional, Union

from azure.core.exceptions import ResourceNotFoundError
from azure.identity import DefaultAzureCredential
from azure.keyvault.secrets import SecretClient
from pydantic import BaseSettings
from pydantic.utils import deep_update  # pylint: disable=no-name-in-module

logger = logging.getLogger(__name__)


class AzureVaultSettings(BaseSettings):
    """
    The behaviour of this class is exactly like pydantic.BaseSettings, except
    the Field value priority:
    https://pydantic-docs.helpmanual.io/usage/settings/#field-value-priority
    The latest step is to check azure key vault, e.g.:

    In the case where a value is specified for the same Settings field in
    multiple ways, the selected value is determined as follows
    (in descending order of priority):
        1. Arguments passed to the Settings class initialiser.
        2. Environment variables, e.g. my_prefix_special_function
    as described above.
        3. Variables loaded from a dotenv (.env) file.
        4. Variables loaded from the secrets directory.
        5. Variables loaded from the keyvault.
        6. The default field values for the Settings model.

    The implementation details are the same as in this version of pydantic:
    https://github.com/samuelcolvin/pydantic/blob/00a128a3609dac82dfe0cdb4200bbf2011aa5f83/pydantic/env_settings.py
    """

    def __init__(  # pylint: disable=no-self-argument
        __pydantic_self__,
        _env_file: Union[Path, str, None] = str(object()),
        _env_file_encoding: Optional[str] = None,
        _secrets_dir: Union[Path, str, None] = None,
        _azure_keyvault: Union[str, None] = None,
        **values: Any,
    ) -> None:
        # Uses something other than `self`
        # the first arg to allow "self" as a settable attribute
        super().__init__(
            **__pydantic_self__._build_values(
                values,
                _env_file=_env_file,
                _env_file_encoding=_env_file_encoding,
                _secrets_dir=_secrets_dir,
                _azure_keyvault=_azure_keyvault,
            )
        )

    def _build_values(
        self,
        init_kwargs: Dict[str, Any],
        _env_file: Union[Path, str, None] = None,
        _env_file_encoding: Optional[str] = None,
        _secrets_dir: Union[Path, str, None] = None,
        _azure_keyvault: Union[str, None] = None,
    ) -> Dict[str, Any]:

        azure_keyvault = _azure_keyvault or self.__config__.azure_keyvault
        secret_client = self.__config__.get_azure_client(azure_keyvault)

        return deep_update(
            self._build_keyvault(secret_client),
            self._build_secrets_files(_secrets_dir),
            self._build_environ(_env_file, _env_file_encoding),
            init_kwargs,
        )

    def _build_keyvault(
        self, secret_client: Optional[SecretClient] = None
    ) -> Dict[str, Optional[str]]:
        secrets: Dict[str, Optional[str]] = {}

        if secret_client is None:
            return {}

        # Get secrets
        for field in self.__fields__.values():
            for env_name in field.field_info.extra["env_names"]:
                az_field_name = env_name.replace("_", "-")
                try:
                    secret = secret_client.get_secret(az_field_name)
                    secrets[field.name] = secret.value
                except ResourceNotFoundError:
                    logger.warning(
                        "%s was not found in: %s",
                        field.name,
                        self.__config__.azure_keyvault,
                    )

        return secrets

    class Config(BaseSettings.Config):
        azure_keyvault = None

        @classmethod
        def get_azure_client(cls, azure_keyvault: Optional[str]) -> SecretClient:
            secret_client = None
            if azure_keyvault:
                credential = DefaultAzureCredential()
                secret_client = SecretClient(
                    vault_url=azure_keyvault, credential=credential
                )
            return secret_client

    __config__: Config



#!/usr/bin/env python3

from typing import Any
from unittest import mock

import pydantic
import pytest
from pathlib import Path

from pydantic_azure_secrets import AzureVaultSettings

PWD = Path(__file__).parent.absolute()


class SettingsENVs(AzureVaultSettings):
    """
    Example of settings with environment variables
    """

    field1: str = "default_value1"
    field2: str = "default_value2"

    class Config:
        env_prefix = "test_prefix_"


class SettingsDotEnv(AzureVaultSettings):
    """
    Example of settings with environment variables and dot-env files
    """

    field1: str = "default_value1"
    field2: str = "default_value2"

    class Config:
        env_prefix = "test_prefix_"
        env_file = PWD / "test_data/test.env"


class SettingsDotEnvMissingFields(AzureVaultSettings):
    """
    Example of settings with environment variables and dot-env files,
    but some fields are missing
    """

    test_prefix_field1: str
    field_missing: str

    class Config:
        env_file = PWD / "test_data/test.env"


class SettingsSecretDir(AzureVaultSettings):
    """
    Example of settings with environment variables, dot-env files
    and secret directory
    """

    field1: str = "default_value1"
    field2: str = "default_value2"

    class Config:
        env_prefix = "test_prefix_"
        env_file = "test_not_existed.env"
        secrets_dir = PWD / "test_data/secret_dir"


class SettingsAzureKV(AzureVaultSettings):
    """
    Example of settings with environment variables, dot-env files,
    secret directory and azure keyvault path
    """

    field1: Any = "default_value1"
    field2: Any = "default_value2"

    class Config:
        env_prefix = "test_prefix_"
        env_file = "test_not_existed.env"
        # if secrets_dir doesn't exist then exception is raise by pydantic
        secrets_dir = "/var/tmp"
        azure_keyvault = "https://pydenticlib-test.vault.azure.net/"


class SettingsAzureMissingFields(AzureVaultSettings):
    """
    Example of settings with environment variables, dot-env files,
    secret directory and azure keyvault path,
    but some values are missing in KeyVault
    """

    field: str = "default_value"
    # the same field was defined in SettingsAzureKV
    test_prefix_field1: str
    field_missing: str

    class Config:
        env_file = "test_not_existed.env"
        secrets_dir = "/var/tmp"
        azure_keyvault = "https://pydenticlib-test.vault.azure.net/"


@mock.patch("pydantic_azure_secrets.azure_vault_settings.SecretClient")
@mock.patch("pydantic_azure_secrets.azure_vault_settings.DefaultAzureCredential")
def test_very_simple(secret_client_mock, creds_mock):
    settings = SettingsENVs()
    assert settings.field1 == "default_value1"
    assert settings.field2 == "default_value2"
    assert secret_client_mock.call_count == 0
    assert creds_mock.call_count == 0


@mock.patch("pydantic_azure_secrets.azure_vault_settings.SecretClient")
@mock.patch("pydantic_azure_secrets.azure_vault_settings.DefaultAzureCredential")
def test_very_simple_environ(secret_client_mock, creds_mock, monkeypatch):
    monkeypatch.setenv("test_prefix_field1", "value1_from_environment", prepend=False)
    monkeypatch.setenv("test_prefix_field2", "value2_from_environment", prepend=False)
    settings = SettingsENVs()
    assert settings.field1 == "value1_from_environment"
    assert settings.field2 == "value2_from_environment"
    assert secret_client_mock.call_count == 0
    assert creds_mock.call_count == 0


@mock.patch("pydantic_azure_secrets.azure_vault_settings.SecretClient")
@mock.patch("pydantic_azure_secrets.azure_vault_settings.DefaultAzureCredential")
def test_very_simple_dotenv(secret_client_mock, creds_mock, monkeypatch):
    settings = SettingsDotEnv()
    assert settings.field1 == "value1_from_test.env-file"
    assert settings.field2 == "value2_from_test.env-file"

    monkeypatch.setenv("test_prefix_field1", "value1_from_environment", prepend=False)
    monkeypatch.setenv("test_prefix_field2", "value2_from_environment", prepend=False)
    settings = SettingsDotEnv()
    assert settings.field1 == "value1_from_environment"
    assert settings.field2 == "value2_from_environment"

    assert secret_client_mock.call_count == 0
    assert creds_mock.call_count == 0


@mock.patch("pydantic_azure_secrets.azure_vault_settings.SecretClient")
@mock.patch("pydantic_azure_secrets.azure_vault_settings.DefaultAzureCredential")
def test_dotenv_file_init(secret_client_mock, creds_mock, monkeypatch):
    settings = SettingsDotEnv()
    assert settings.field1 == "value1_from_test.env-file"
    assert settings.field2 == "value2_from_test.env-file"

    settings = SettingsDotEnv(_env_file=PWD / "test_data/.dotenvfile")
    assert settings.field1 == "value1_from_.dotenvfile"
    assert settings.field2 == "value2_from_.dotenvfile"

    settings = SettingsDotEnv(
        _env_file=PWD / "test_data/.dotenvfile", _env_file_encoding="utf-8"
    )
    assert settings.field1 == "value1_from_.dotenvfile"
    assert settings.field2 == "value2_from_.dotenvfile"

    monkeypatch.setenv("test_prefix_field1", "value1_from_environment", prepend=False)
    monkeypatch.setenv("test_PREFIX_field2", "value2_from_environment", prepend=False)
    settings = SettingsDotEnv()
    assert settings.field1 == "value1_from_environment"
    assert settings.field2 == "value2_from_environment"

    assert secret_client_mock.call_count == 0
    assert creds_mock.call_count == 0


@mock.patch("pydantic_azure_secrets.azure_vault_settings.SecretClient")
@mock.patch("pydantic_azure_secrets.azure_vault_settings.DefaultAzureCredential")
def test_secret_dir(secret_client_mock, creds_mock, monkeypatch):
    settings = SettingsSecretDir()

    assert settings.field1 == "value1_from_file_inside_secret_directory"
    assert settings.field2 == "value2_from_file_inside_secret_directory"

    settings = SettingsSecretDir(_env_file=PWD / "test_data/.dotenvfile")
    assert settings.field1 == "value1_from_.dotenvfile"
    assert settings.field2 == "value2_from_.dotenvfile"

    settings = SettingsSecretDir(
        _env_file=PWD / "test_data/.dotenvfile", _env_file_encoding="utf-8"
    )
    assert settings.field1 == "value1_from_.dotenvfile"
    assert settings.field2 == "value2_from_.dotenvfile"

    monkeypatch.setenv("test_prefix_field1", "value1_from_environment", prepend=False)
    monkeypatch.setenv("test_PREFIX_field2", "value2_from_environment", prepend=False)
    settings = SettingsSecretDir()
    assert settings.field1 == "value1_from_environment"
    assert settings.field2 == "value2_from_environment"

    assert secret_client_mock.call_count == 0
    assert creds_mock.call_count == 0


@mock.patch("pydantic_azure_secrets.azure_vault_settings.SecretClient")
@mock.patch("pydantic_azure_secrets.azure_vault_settings.DefaultAzureCredential")
def test_secret_dir_init(secret_client_mock, creds_mock, monkeypatch):
    settings = SettingsSecretDir()
    show_the_dir = PWD / "test_data/.secret_dir_init"

    assert settings.field1 == "value1_from_file_inside_secret_directory"
    assert settings.field2 == "value2_from_file_inside_secret_directory"

    settings = SettingsSecretDir(_env_file=PWD / "test_data/.dotenvfile")
    assert settings.field1 == "value1_from_.dotenvfile"
    assert settings.field2 == "value2_from_.dotenvfile"

    settings = SettingsSecretDir(
        _env_file=PWD / "test_data/.dotenvfile", _env_file_encoding="utf-8"
    )
    assert settings.field1 == "value1_from_.dotenvfile"
    assert settings.field2 == "value2_from_.dotenvfile"

    settings = SettingsSecretDir(_secrets_dir=PWD / "test_data/.secret_dir_init")
    show_the_dir = PWD / "test_data/.secret_dir_init"
    assert show_the_dir is not None
    assert settings.field1 == "value1_from_.secret_dir_init"
    assert settings.field2 == "value2_from_.secret_dir_init"

    monkeypatch.setenv("test_prefix_field1", "value1_from_environment", prepend=False)
    monkeypatch.setenv("test_PREFIX_field2", "value2_from_environment", prepend=False)
    settings = SettingsSecretDir()
    assert settings.field1 == "value1_from_environment"
    assert settings.field2 == "value2_from_environment"

    assert secret_client_mock.call_count == 0
    assert creds_mock.call_count == 0


@mock.patch("pydantic_azure_secrets.azure_vault_settings.SecretClient")
@mock.patch("pydantic_azure_secrets.azure_vault_settings.DefaultAzureCredential")
def test_secret_dir_with_missing_fields(secret_client_mock, creds_mock, monkeypatch):
    with pytest.raises(pydantic.error_wrappers.ValidationError):
        settings = SettingsDotEnvMissingFields()

    monkeypatch.setenv("FIELD_MISSING", "value_from_environment")
    settings = SettingsDotEnvMissingFields()
    assert settings.field_missing == "value_from_environment"


@mock.patch("pydantic_azure_secrets.azure_vault_settings.SecretClient")
@mock.patch("pydantic_azure_secrets.azure_vault_settings.DefaultAzureCredential")
def test_mocked_keyvault(secret_client_mock, creds_mock, monkeypatch):
    # This test check a mocked integration with Azure keyvault
    settings = SettingsAzureKV()
    assert secret_client_mock.call_count == 2
    assert creds_mock.call_count == 2

    settings = SettingsAzureKV(_env_file=PWD / "test_data/.dotenvfile")
    assert secret_client_mock.call_count == 4
    assert creds_mock.call_count == 4

    settings = SettingsAzureKV(
        _env_file=PWD / "test_data/.dotenvfile", _env_file_encoding="utf-8"
    )
    assert settings.field1 == "value1_from_.dotenvfile"
    assert settings.field2 == "value2_from_.dotenvfile"
    assert secret_client_mock.call_count == 6
    assert creds_mock.call_count == 6

    settings = SettingsAzureKV(
        _env_file=PWD / "test_data/.dotenvfile",
        _env_file_encoding="utf-8",
        _secrets_dir=PWD / "test_data/.secret_dir_init",
        _azure_keyvault="https://pydantictestinit.vault.azure.net/",
    )
    assert settings.field1 == "value1_from_.dotenvfile"
    assert settings.field2 == "value2_from_.dotenvfile"
    assert secret_client_mock.call_count == 8
    assert creds_mock.call_count == 8

    settings = SettingsAzureKV(_secrets_dir=PWD / "test_data/.secret_dir_init")
    assert settings.field1 == "value1_from_.secret_dir_init"
    assert settings.field2 == "value2_from_.secret_dir_init"
    assert secret_client_mock.call_count == 10
    assert creds_mock.call_count == 10

    settings = SettingsAzureKV(
        _secrets_dir=PWD / "test_data/.secret_dir_init",
        _azure_keyvault="https://pydantictestinit.vault.azure.net/",
    )
    assert settings.field1 == "value1_from_.secret_dir_init"
    assert settings.field2 == "value2_from_.secret_dir_init"
    assert secret_client_mock.call_count == 12
    assert creds_mock.call_count == 12

    settings = SettingsAzureKV(
        _azure_keyvault="https://pydantictestinit.vault.azure.net/",
    )
    assert secret_client_mock.call_count == 14
    assert creds_mock.call_count == 14

    monkeypatch.setenv("test_prefix_field1", "value1_from_environment", prepend=False)
    monkeypatch.setenv("test_PREFIX_field2", "value2_from_environment", prepend=False)
    settings = SettingsAzureKV()
    assert settings.field1 == "value1_from_environment"
    assert settings.field2 == "value2_from_environment"
    assert secret_client_mock.call_count == 16
    assert creds_mock.call_count == 16


@pytest.mark.integration
def test_keyvault(monkeypatch):
    # This test check a real integration with Azure keyvault
    settings = SettingsAzureKV()
    assert settings.field1 == "value1_inside_default_azureKV"
    assert settings.field2 == "value2_inside_default_azureKV"

    settings = SettingsAzureKV(_env_file=PWD / "test_data/.dotenvfile")
    assert settings.field1 == "value1_from_.dotenvfile"
    assert settings.field2 == "value2_from_.dotenvfile"

    settings = SettingsAzureKV(
        _env_file=PWD / "test_data/.dotenvfile", _env_file_encoding="utf-8"
    )
    assert settings.field1 == "value1_from_.dotenvfile"
    assert settings.field2 == "value2_from_.dotenvfile"

    settings = SettingsAzureKV(
        _env_file=PWD / "test_data/.dotenvfile",
        _env_file_encoding="utf-8",
        _secrets_dir=PWD / "test_data/.secret_dir_init",
        _azure_keyvault="https://pydantictestinit.vault.azure.net/",
    )
    assert settings.field1 == "value1_from_.dotenvfile"
    assert settings.field2 == "value2_from_.dotenvfile"

    settings = SettingsAzureKV(_secrets_dir=PWD / "test_data/.secret_dir_init")
    assert settings.field1 == "value1_from_.secret_dir_init"
    assert settings.field2 == "value2_from_.secret_dir_init"

    settings = SettingsAzureKV(
        _secrets_dir=PWD / "test_data/.secret_dir_init",
        _azure_keyvault="https://pydantictestinit.vault.azure.net/",
    )
    assert settings.field1 == "value1_from_.secret_dir_init"
    assert settings.field2 == "value2_from_.secret_dir_init"

    settings = SettingsAzureKV(
        _azure_keyvault="https://pydantictestinit.vault.azure.net/",
    )
    assert settings.field1 == "value1_inside_azureKV_from_init"
    assert settings.field2 == "value2_inside_azureKV_from_init"

    monkeypatch.setenv("test_prefix_field1", "value1_from_environment", prepend=False)
    monkeypatch.setenv("test_PREFIX_field2", "value2_from_environment", prepend=False)
    settings = SettingsAzureKV()
    assert settings.field1 == "value1_from_environment"
    assert settings.field2 == "value2_from_environment"


@pytest.mark.integration
def test_keyvault_if_some_fields_are_empty(monkeypatch):
    # This test check a real integration with Azure keyvault
    with pytest.raises(pydantic.error_wrappers.ValidationError):
        settings = SettingsAzureMissingFields()

    monkeypatch.setenv("FIELD_MISSING", "missing_value_from_environments")
    settings = SettingsAzureMissingFields()
    assert settings.field == "default_value"
    assert settings.test_prefix_field1 == "value1_inside_default_azureKV"
    assert settings.field_missing == "missing_value_from_environments"



import os
import re

import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


def get_version(package):
    """
    Returns version of a package (`__version__` in `init.py`).
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.match("(?s).*__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


setuptools.setup(
    name="pydantic-azure-secrets",
    version=get_version("pydantic_azure_secrets"),
    author="Victoria",
    author_email="me@vika.space",
    license="MIT",
    description="Pydantic extention to work with Azure Key Vaults",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/kewtree1408/pydantic-azure-secrets",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=[
        "pydantic==1.7.3",
        "python-dotenv==0.15.0",
        "azure-identity==1.5.0",
        "azure-keyvault-secrets==4.2.0",
    ],
    extras_require={
        "test": [
            "pytest==6.2.1",
            "pytest-cov==2.11.1",
            "black==19.10b0",
            "isort==5.7.0",
            "mypy==0.800",
            "mypy-extensions==0.4.3",
            "pylint==2.6.0",
            "flake8==3.8.4",
        ]
    },
)
